from fastapi import FastAPI
from fastapi.openapi.utils import get_openapi

app = FastAPI()

def custom_openapi():

    if app.openapi_schema:
        return app.openapi_schema

    openapi_schema = get_openapi(
        title="Dedupe",
        version="2.0.0",
        routes=app.routes,
    )

    app.openapi_schema = openapi_schema

    return app.openapi_schema

app.openapi = custom_openapi



from oagdedupe.settings import Settings

class BlockSchemesHelper:
    """
    Attributes used to help build SQL queries, which are used to 
    build forward indices. 
    """

    @property
    def block_scheme_mapping(self):
        """
        helper to build column names in query
        """
        mapping = {}
        for attribute in self.settings.other.attributes:
            for scheme,nlist in self.block_schemes:
                for n in nlist:
                    if n:
                        mapping[f"{scheme}_{n}_{attribute}"]=f"{scheme}({attribute},{n})"
                    else:
                        mapping[f"{scheme}_{attribute}"]=f"{scheme}({attribute})"
        return mapping

    @property
    def block_scheme_sql(self):
        """
        helper to build column names in query
        """
        return [
            f"{scheme}({attribute},{n}) as {scheme}_{n}_{attribute}"
            if n
            else f"{scheme}({attribute}) as {scheme}_{attribute}"
            for attribute in self.settings.other.attributes
            for scheme,nlist in self.block_schemes
            for n in nlist
        ]

class BlockSchemes(BlockSchemesHelper):
    """
    Contains all block schemes.
    """
    settings: Settings

    @property
    def block_schemes(self):
        """
        List of tuples containing block scheme name and parameters.
        The block scheme name should correspond to a postgres function
        """
        return [
            ("first_nchars", [2,4,6]),
            ("last_nchars", [2,4,6]),
            ("find_ngrams",[2,4,6]),
            ("acronym", [None]),
            ("exactmatch", [None])
        ]

    @property
    def block_scheme_names(self):
        """
        Convenience property to list all block schemes
        """
        return [
            f"{scheme}_{n}_{attribute}"
            if n
            else f"{scheme}_{attribute}"
            for attribute in self.settings.other.attributes
            for scheme,nlist in self.block_schemes
            for n in nlist
        ]



from abc import ABCMeta, abstractmethod


class BaseBlocker(metaclass=ABCMeta):
    """ Abstract base class for all blockers to inherit
    """

    @abstractmethod
    def get_block_maps(self, df, attributes):
        return


class BaseBlockAlgo(metaclass=ABCMeta):
    """ Abstract base class for all blocking algos to inherit
    """

    @abstractmethod
    def get_block(self):
        return


class BaseDistance(metaclass=ABCMeta):
    """ Abstract base class for all distance configurations to inherit
    """

    @abstractmethod
    def distance(self):
        return

    # @abstractmethod
    # def config(self):
    #     return


class BaseCluster(metaclass=ABCMeta):
    """ Abstract base class for all clustering algos to inherit
    """

    @abstractmethod
    def get_df_cluster(self):
        return



from oagdedupe.settings import Settings

from sqlalchemy import create_engine
from dataclasses import dataclass

def create_functions(settings: Settings):

    engine = create_engine(settings.other.path_database, echo=False)

    engine.execute("""
        CREATE OR REPLACE FUNCTION first_nchars(s text, n integer) RETURNS text
        AS $$
        return s[:n]
        $$ 
        LANGUAGE plpython3u;
    """)

    engine.execute("""
        CREATE OR REPLACE FUNCTION last_nchars(s text, n integer) RETURNS text
        AS $$
        return s[-n:]
        $$ 
        LANGUAGE plpython3u;
    """)

    engine.execute("""
        CREATE OR REPLACE FUNCTION find_ngrams(s text, n integer) RETURNS text[]
        AS $$
        return [s[i:i+n] for i in range(len(s)-n+1)]
        $$ 
        LANGUAGE plpython3u;
    """)

    engine.execute("""
        CREATE OR REPLACE  FUNCTION acronym(s text) RETURNS text
        AS $$
        return "".join(e[0] for e in s.split())
        $$ 
        LANGUAGE plpython3u;
    """)

    engine.execute("""
        CREATE OR REPLACE FUNCTION exactmatch(s text) RETURNS text
        AS $$
        return s
        $$ 
        LANGUAGE plpython3u;
    """)

    engine.execute("""
        DROP FUNCTION IF EXISTS unnest_2d_1d(ANYARRAY);
        CREATE OR REPLACE FUNCTION unnest_2d_1d(ANYARRAY, OUT a ANYARRAY)
        RETURNS SETOF ANYARRAY
        LANGUAGE plpgsql IMMUTABLE STRICT AS
        $func$
        BEGIN
        FOREACH a SLICE 1 IN ARRAY $1 LOOP
            RETURN NEXT;
        END LOOP;
        END
        $func$;
    """)

    engine.execute("""
        CREATE OR REPLACE FUNCTION combinations(arr integer[]) RETURNS integer[]
        AS $$
        def combinations(iterable, r):
            # combinations('ABCD', 2) --> AB AC AD BC BD CD
            # combinations(range(4), 3) --> 012 013 023 123
            pool = tuple(iterable)
            n = len(pool)
            if r > n:
                return
            indices = list(range(r))
            yield tuple(pool[i] for i in indices)
            while True:
                for i in reversed(range(r)):
                    if indices[i] != i + n - r:
                        break
                else:
                    return
                indices[i] += 1
                for j in range(i+1, r):
                    indices[j] = indices[j-1] + 1
                yield tuple(pool[i] for i in indices)
        return [[s[0],s[1]] for s in combinations(arr,2)]
        $$ 
        LANGUAGE plpython3u;
    """)







def recordlinkage(f):
    def wrapper(*args, **kwargs):
        self = args[0]
        if self.settings.other.dedupe == False:
            kwargs["rl"] = "_link"
            return f(*args, **kwargs)
        else:
            return f(*args, **kwargs)
    return wrapper

def recordlinkage_both(f):
    def wrapper(*args, **kwargs):
        self = args[0]
        out1 = f(*args, **kwargs)
        if self.settings.other.dedupe == False:
            kwargs["rl"] = "_link"
            out2 = f(*args, **kwargs)
            return out1,out2
        return out1
        
    return wrapper

def recordlinkage_repeat(f):
    def wrapper(*args, **kwargs):
        self = args[0]
        f(*args, **kwargs)
        if self.settings.other.dedupe == False:
            kwargs["rl"] = "_link"
            f(*args, **kwargs)
    return wrapper


from oagdedupe.settings import Settings
from oagdedupe.block.schemes import BlockSchemes
from oagdedupe.db.database import Engine
from oagdedupe import utils as du

from dataclasses import dataclass
import logging
import pandas as pd

@dataclass
class Blocker(BlockSchemes, Engine):
    """
    Used to build forward indices. A forward index 
    is a table where rows are entities, columns are block schemes,
    and values contain signatures.

    Attributes
    ----------
    settings : Settings
    """
    settings: Settings

    def query_blocks(self, table, columns):
        """
        Builds SQL query used to build forward indices. 

        Parameters
        ----------
        table : str
        columns : List[str]

        Returns
        ----------
        str
        """
        
        return f"""
            DROP TABLE IF EXISTS {self.settings.other.db_schema}.blocks_{table};
            
            CREATE TABLE {self.settings.other.db_schema}.blocks_{table} as (
                SELECT 
                    _index,
                    {", ".join(columns)}
                FROM {self.settings.other.db_schema}.{table}
            );
        """
        
    @du.recordlinkage_repeat
    def add_scheme(self, table, col, exists, rl=""):
        """
        check if column is in exists
        if not, add to blocks_{tablle}
        """

        if col in exists:
            return

        if "ngrams" in col:
            coltype = "text[]"
        else:
            coltype = "text"

        self.engine.execute(f"""
            ALTER TABLE {self.settings.other.db_schema}.blocks_{table}{rl}
            ADD COLUMN IF NOT EXISTS {col} {coltype}
        """)

        self.engine.execute(f"""
            UPDATE {self.settings.other.db_schema}.blocks_{table}{rl} AS t1
            SET {col} = t2.{col}
            FROM (
                SELECT _index, {self.block_scheme_mapping[col]} as {col}
                FROM {self.settings.other.db_schema}.{table}{rl}
            ) t2
            WHERE t1._index = t2._index;
        """)
        self.engine.dispose()
        return

    @du.recordlinkage_repeat
    def build_forward_indices(self, rl=""):
        """
        Executes SQL queries to build forward indices for train datasets
        """
    
        logging.info(f"Building forward indices")
        self.engine.execute(self.query_blocks(
            table=f"train{rl}",
            columns=self.block_scheme_sql
        ))

    @du.recordlinkage_repeat
    def init_forward_index_full(self, rl=""):
    
        self.engine.execute(f"""
            DROP TABLE IF EXISTS {self.settings.other.db_schema}.blocks_df{rl};
            
            CREATE TABLE {self.settings.other.db_schema}.blocks_df{rl} as (
                SELECT 
                    _index
                FROM {self.settings.other.db_schema}.df{rl}
            );
        """)

    def build_forward_indices_full(self, columns):
        """
        Executes SQL queries to build forward indices on full data.

        Parameters
        ----------
        columns : List[str]
            block schemes to include in forward index
        """
        for col in columns:
            exists = pd.read_sql(
                f"SELECT * FROM {self.settings.other.db_schema}.blocks_df LIMIT 1", 
                con=self.engine
            ).columns
        
            self.add_scheme(table="df", col=col, exists=exists)



from oagdedupe.labelstudio.lsapi import LabelStudioAPI
from oagdedupe.db.database import Engine
from oagdedupe.settings import Settings
from oagdedupe.api import Dedupe

from dataclasses import dataclass
from typing import List
from modAL.models import ActiveLearner
from modAL.uncertainty import uncertainty_sampling
from sklearn.ensemble import RandomForestClassifier
import requests
import joblib
import pandas as pd
import logging
from sqlalchemy import select, update

def url_checker(url):
    try:
        get = requests.get(url)
        if get.status_code == 200:
            return True
        else:
            return False
    except Exception as e:
        return False
    
class Projects:
    """
    Fetch Project if it exists; else create new one
    """

    def _check_project_exists(self):
        for project in self.lsapi.list_projects():
            if project.title == self.title:
                return project
        return False

    def _setup_project(self):
        project = self._check_project_exists()
        if not project:
            project = self.lsapi.create_project(
                title=self.title, description=self.description
            )
        return project
    
    def initialize_project(self):
        """
        check for new project; create if not exist
        """
        self.title = self.settings.name
        self.description = self.settings.other.label_studio.description
        self.project = self._setup_project()
        if not self.lsapi.get_webhooks():
            self.lsapi.post_webhook(project_id=self.project.id)

class Tasks:
    """
    Pushes new annotations to label studio if needed.
    """

    def _update_train(self, newlabels):
        indices = set(newlabels["_index_l"]).union(set(newlabels["_index_r"]))
        with self.api.init.Session() as session:
            stmt = (
                update(self.api.init.Train).
                where(self.api.init.Train._index.in_(indices)).
                values(labelled=True)
            )
            session.execute(stmt)
            session.commit()
    
    def _update_labels_and_train_tables(self, newlabels):
        self.api.orm._update_table(newlabels, self.api.init.Labels())
        self._update_train(newlabels)
        return True

    def _get_new_labels(self):
        return self.lsapi.get_new_labels(project_id=self.project.id)

    def _get_learning_samples(self, n_instances=5):
        
        logging.info("getting distances")
        distances = self.api.orm.get_distances()

        logging.info("getting active learning samples")
        sample_idx, _ = self.clf.query(
            distances[self.settings.other.attributes].values, 
            n_instances=n_instances
        )

        return distances.loc[
            sample_idx,
            self.api.orm.get_compare_cols()
        ]
            
    def _post_tasks(self):
        learning_samples = self._get_learning_samples(n_instances=10)
        self.lsapi.post_tasks(df=learning_samples, project_id=self.project.id)
        return

    def generate_new_samples(self):

        tasklist = self.lsapi.get_tasks(project_id=self.project.id)

        if tasklist.total == 0:
            self._post_tasks()
            return
        
        if tasklist.n_incomplete < 5:
            
            # get new annotations and update
            newlabels = self._get_new_labels()
            if len(newlabels) > 0:
                self._update_labels_and_train_tables(newlabels)
            else:
                return

            # learn new block conjunctions
            self.api.initialize(reset=False, resample=True)
            
            # re-train model
            self._train()

            # post new active learning samples to label studio
            self._post_tasks()

@dataclass
class Model(Tasks, Projects, Engine):
    settings:Settings

    def __post_init__(self):
        self.api =  Dedupe(settings=self.settings)
        self.lsapi = LabelStudioAPI(settings=self.settings)

    def initialize_learner(self):
        """
        initialize active learner
        """
        if self.settings.other.path_model.is_file():
            self.estimator = joblib.load(self.settings.other.path_model)
            logging.info(f"reading model: {self.settings.other.path_model}")
        else:
            self.estimator = RandomForestClassifier(self.settings.other.cpus)

        self.clf = ActiveLearner(
            estimator=self.estimator, query_strategy=uncertainty_sampling
        )
        self._train()
    
    def _train(self):
        labels=self.api.orm.get_labels()
        self.clf.teach(labels[self.settings.other.attributes].values, labels["label"].values)
        


from oagdedupe.fastapi import fapi
from oagdedupe.fastapi import app
from oagdedupe.labelstudio import lsapi
from oagdedupe.settings import get_settings_from_env
from typing import Union

import time
import joblib
import uvicorn
import logging

root = logging.getLogger()
root.setLevel(logging.DEBUG)

settings = get_settings_from_env()
assert settings.other is not None
while fapi.url_checker(settings.other.label_studio.url) == False:
    logging.info("waiting for label studio...")
    time.sleep(3)

m = fapi.Model(settings=settings)
m.initialize_learner()


@app.on_event("startup")
async def startup():
    m.initialize_project()
    m.generate_new_samples()


@app.get("/predict")
async def predict():

    logging.info(f"save model to {settings.other.path_model}")
    joblib.dump(m.clf.estimator, settings.other.path_model)

    return dict(
        {
            "predict_proba": m.clf.predict_proba(
                m.api.orm.get_full_distances()[settings.other.attributes].values
            )[:,1].tolist(),
            "predict": m.clf.predict(
                m.api.orm.get_full_distances()[settings.other.attributes].values
            ).tolist(),
        }
    )


@app.post("/payload")
async def payload():
    m.generate_new_samples()


if __name__ == "__main__":
    uvicorn.run(
        app, host=settings.other.fast_api.host.split("/")[-1], port=settings.other.fast_api.port
    )



from oagdedupe.base import BaseDistance
from oagdedupe.db.database import DatabaseORM
from oagdedupe.settings import Settings

from dataclasses import dataclass
from jellyfish import jaro_winkler_similarity
import ray
import numpy as np
import pandas as pd
import logging

@ray.remote
def ray_distance(pairs):
    return [
        jaro_winkler_similarity(pair[0], pair[1])
        for pair in pairs
    ]

@dataclass
class RayAllJaro(BaseDistance, DatabaseORM):
    """
    Interface to compute distance between comparison pairs along 
    common attributes.
    """
    settings: Settings

    def __post_init__(self):
        self.orm = DatabaseORM(settings=self.settings)

    def distance(self, pairs):
        """
        Parameters
        ----------
        pairs: np.array
            Nx2 array of comparison pairs
        """
        chunks = self.get_chunks(pairs, 10000)
        res = ray.get([
            ray_distance.remote(chunk)
            for chunk in chunks
        ])
        return [
            x for sublist in res for x in sublist
        ]

    def get_chunks(self, lst, n):
        """Yield successive n-sized chunks from lst."""
        for i in range(0, len(lst), n):
            yield lst[i:i + n]

    def get_distmat(self, table) -> np.array:
        """
        get comparison attributes from table then compute distances
        for each pair
        
        Parameters
        ----------
        table: sqlalchemy.orm.decl_api.DeclarativeMeta

        Returns
        ----------
        pd.DataFrame
        """

        comps = self.orm.get_comparison_attributes(table)
        
        logging.info(
            f"making {len(comps)} comparions for table: {table.__tablename__}")

        distances = {
            attribute:self.distance(
                comps[[f"{attribute}_l",f"{attribute}_r"]].values)
            for attribute in self.settings.other.attributes
        }

        return pd.concat([
            comps,
            pd.DataFrame(distances)
        ], axis=1)

    def save_distances(self, table, newtable):
        """
        saves distances to newtable
        """

        distances = self.get_distmat(table=table)
        
        # reset table
        self.orm.engine.execute(f"""
            TRUNCATE TABLE {self.settings.other.db_schema}.{newtable.__tablename__};
        """)

        # insert
        self.orm.bulk_insert(
            df=distances, to_table=newtable
        )



from oagdedupe.distance.string import RayAllJaro
from oagdedupe.cluster.cluster import ConnectedComponents
from oagdedupe.settings import Settings
from oagdedupe.block import Blocker, Conjunctions
from oagdedupe.db.initialize import Initialize
from oagdedupe.db.database import DatabaseORM

import requests
import json
from abc import ABCMeta, abstractmethod
from typing import List, Optional, Tuple
from dataclasses import dataclass
from functools import cached_property
import pandas as pd
import numpy as np
import ray
from sqlalchemy import create_engine
import logging

root = logging.getLogger()
root.setLevel(logging.DEBUG)


class BaseModel(metaclass=ABCMeta):
    """Abstract base class from which all model classes inherit.
    All descendent classes must implement predict, train, and candidates methods.
    """

    """project settings"""

    @abstractmethod
    def predict(self):
        return

    @abstractmethod
    def fit_blocks(self):
        return

    @abstractmethod
    def fit_model(self):
        return

    @abstractmethod
    def initialize(self):
        return

    @cached_property
    def engine(self):
        return create_engine(self.settings.other.path_database)

    @cached_property
    def init(self):
        return Initialize(settings=self.settings)

    @cached_property
    def orm(self):
        return DatabaseORM(settings=self.settings)

    @cached_property
    def blocker(self):
        return Blocker(settings=self.settings)

    @cached_property
    def cover(self):
        return Conjunctions(settings=self.settings)
    
    @cached_property
    def distance(self):
        return RayAllJaro(settings=self.settings)

@dataclass
class FitModel:

    def fit_model(self) -> Tuple[np.array, np.array, np.array]:
        """learn p(match)"""

        # get predictions
        results = json.loads(
            requests.get(f"{self.settings.other.fast_api.url}/predict").content
        )

        return (
            self.orm.get_full_comparison_indices().values,
            np.array(results["predict_proba"]),
            np.array(results["predict"])
        )

@dataclass
class Dedupe(FitModel, BaseModel):
    """General dedupe block, inherits from BaseModel."""
    settings:Settings

    def __post_init__(self):
        self.settings.sync()
        if (self.settings.other.cpus > 1) & (not ray.is_initialized()):
            ray.init(num_cpus=self.settings.other.cpus)
    
    def predict(self) -> pd.DataFrame:
        """get clusters of matches and return cluster IDs"""

        idxmat, scores, y = self.fit_model()
        self.cluster = ConnectedComponents(settings=self.settings)
        logging.info("get clusters")
        return self.cluster.get_df_cluster(
            matches=idxmat[y == 1].astype(int), scores=scores[y == 1]
        )

    def fit_blocks(self, n_covered=2_000_000):

        # fit block scheme conjunctions to full data
        self.blocker.init_forward_index_full()
        self.cover.save_best(
            table="blocks_df", newtable="full_comparisons", n_covered=n_covered
        )

        # get distances
        self.distance.save_distances(
            table=self.orm.FullComparisons,
            newtable=self.orm.FullDistances
        )

    def initialize(
        self, 
        df=None, 
        reset=True, 
        resample=False, 
        n_covered=500
        ):
        """learn p(match)"""

        self.init.setup(df=df, reset=reset, resample=resample)
        
        self.blocker.build_forward_indices()
        self.cover.save_best(
            table="blocks_train", newtable="comparisons", n_covered=n_covered
        )

        logging.info("get distance matrix")
        self.distance.save_distances(
            table=self.orm.Comparisons,
            newtable=self.orm.Distances
        )

        
@dataclass
class RecordLinkage(FitModel, BaseModel):
    """General dedupe block, inherits from BaseModel."""
    settings:Settings

    def __post_init__(self):
        self.settings.sync()
        if (self.settings.other.cpus > 1) & (not ray.is_initialized()):
            ray.init(num_cpus=self.settings.other.cpus)
    
    def predict(self) -> pd.DataFrame:
        """get clusters of matches and return cluster IDs"""

        idxmat, scores, y = self.fit_model()
        self.cluster = ConnectedComponents(settings=self.settings)
        logging.info("get clusters")
        return self.cluster.get_df_cluster(
            idxmat[y == 1].astype(int), scores[y == 1]
        )

    def fit_blocks(self, n_covered=500_000):

        # fit block scheme conjunctions to full data
        self.blocker.init_forward_index_full()
        self.cover.save_best(
            table="blocks_df", newtable="full_comparisons", n_covered=n_covered
        )

        # get distances
        self.distance.save_distances(
            table=self.orm.FullComparisons,
            newtable=self.orm.FullDistances
        )

    def initialize(
        self, 
        df=None, 
        df2=None,
        reset=True, 
        resample=False, 
        n_covered=500
        ):
        """learn p(match)"""

        self.init.setup(df=df, df2=df2, reset=reset, resample=resample)
        
        self.blocker.build_forward_indices()
        self.cover.save_best(
            table="blocks_train", newtable="comparisons", n_covered=n_covered
        )

        logging.info("get distance matrix")
        self.distance.save_distances(
            table=self.orm.Comparisons,
            newtable=self.orm.Distances
        )

        
        
        





from oagdedupe import utils as du
from oagdedupe.settings import Settings

from dataclasses import dataclass
from functools import cached_property
from sqlalchemy import Column, String, Integer, Boolean
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy import MetaData, create_engine
from sqlalchemy.schema import CreateSchema

class TablesRecordLinkage:

    @cached_property        
    def maindf_link(self):
        return type('df_link', (self.Attributes, self.Base), {
                "__tablename__":"df_link",
                "_index":Column(Integer, primary_key=True)
            }
        )

    @cached_property    
    def Neg_link(self):
        return type('neg_link', (self.Attributes, self.Base), {
                "__tablename__":"neg_link",
                "_index":Column(Integer, primary_key=True),
                "labelled":Column(Boolean)
            }
        )

    @cached_property    
    def Unlabelled_link(self):
        return type('unlabelled_link', (self.Attributes, self.Base), {
                "__tablename__":"unlabelled_link",
                "_index":Column(Integer, primary_key=True),
                "labelled":Column(Boolean)
            }
        )

    @cached_property    
    def Train_link(self):
        return type('train_link', (self.Attributes, self.Base), {
                "__tablename__":"train_link",
                "_index":Column(Integer, primary_key=True),
                "labelled":Column(Boolean)
            }
        )


@dataclass
class Tables(TablesRecordLinkage):
    """
    Factory to create sql alchemy "Declarative Table" 
    (https://docs.sqlalchemy.org/en/14/orm/declarative_tables.html)

    We use builtin function type() to generate declarative table dynamically
    since attribute names are variable
    """
    settings: Settings

    def setup_dynamic_declarative_mapping(self):
        """
        initializes declarative tables
        """
        return (
            self.maindf, 
            self.Pos, 
            self.Neg,
            self.Unlabelled,
            self.Train, 

            self.maindf_link, 
            self.Neg_link,
            self.Unlabelled_link,
            self.Train_link, 

            self.Labels,
            self.Distances,
            self.FullDistances,
            self.Comparisons,
            self.FullComparisons,
            self.Clusters
        )

    @cached_property
    def metadata_obj(self):
        """
        A sqlalchemy MetaData collection that stores Tables
        """
        return MetaData(schema=self.settings.other.db_schema)

    @cached_property
    def Base(self):
        """
        All Table objects declared by subclasses of this Base share a 
        common metadata; that is they are all part of the same 
        Metadata collection
        """
        return declarative_base(metadata=self.metadata_obj)

    @cached_property
    def engine(self):
        return create_engine(self.settings.other.path_database)

    @cached_property
    def Session(self):
        return sessionmaker(bind=self.engine)


    @property
    def Attributes(self):
        return type('Attributes', (object,), {
                k:Column(String)
                for k in self.settings.other.attributes
            }
        )

    @property
    def AttributeComparisons(self):
        return type('AttributeComparisons', (object,), {
                **{
                    f"{k}_l":Column(String)
                    for k in self.settings.other.attributes
                },
                **{
                    f"{k}_r":Column(String)
                    for k in self.settings.other.attributes
                },
            }
        )

    @cached_property        
    def maindf(self):
        return type('df', (self.Attributes, self.Base), {
                "__tablename__":"df",
                "_index":Column(Integer, primary_key=True)
            }
        )

    @cached_property            
    def Pos(self):
        return type('pos', (self.Attributes, self.Base), {
                "__tablename__":"pos",
                "_index":Column(Integer, primary_key=True),
                "labelled":Column(Boolean)
            }
        )

    @cached_property    
    def Neg(self):
        return type('neg', (self.Attributes, self.Base), {
                "__tablename__":"neg",
                "_index":Column(Integer, primary_key=True),
                "labelled":Column(Boolean)
            }
        )


    @cached_property    
    def Unlabelled(self):
        return type('unlabelled', (self.Attributes, self.Base), {
                "__tablename__":"unlabelled",
                "_index":Column(Integer, primary_key=True),
                "labelled":Column(Boolean)
            }
        )

    @cached_property    
    def Train(self):
        return type('train', (self.Attributes, self.Base), {
                "__tablename__":"train",
                "_index":Column(Integer, primary_key=True),
                "labelled":Column(Boolean)
            }
        )

    @cached_property    
    def Labels(self):
        return type('labels', (self.Attributes, self.AttributeComparisons, self.Base), {
                "__tablename__":"labels",
                "_index_l":Column(Integer, primary_key=True),
                "_index_r":Column(Integer, primary_key=True),
                "label":Column(Integer)
            }
        )

    @cached_property    
    def Distances(self):
        return type('distances', (self.Attributes, self.AttributeComparisons, self.Base), {
                "__tablename__":"distances",
                "_index_l":Column(Integer, primary_key=True),
                "_index_r":Column(Integer, primary_key=True),
                "label":Column(Integer)
            }
        )

    @cached_property    
    def FullDistances(self):
        return type('full_distances', (self.Attributes, self.AttributeComparisons, self.Base), {
                "__tablename__":"full_distances",
                "_index_l":Column(Integer, primary_key=True),
                "_index_r":Column(Integer, primary_key=True),
                "label":Column(Integer)
            }
        )

    @cached_property    
    def Comparisons(self):
        return type('comparisons', (self.Base, ), {
                "__tablename__":"comparisons",
                "_index_l":Column(Integer, primary_key=True),
                "_index_r":Column(Integer, primary_key=True)
            }
        )

    @cached_property    
    def FullComparisons(self):
        return type('full_comparisons', (self.Base, ), {
                "__tablename__":"full_comparisons",
                "_index_l":Column(Integer, primary_key=True),
                "_index_r":Column(Integer, primary_key=True)
            }
        )

    @cached_property    
    def Clusters(self):
        return type('clusters', (self.Base, ), {
                "__tablename__":"clusters",
                "_cluster_key":Column(Integer, primary_key=True, autoincrement=True),
                "cluster":Column(Integer),
                "_type":Column(Boolean),
                "_index":Column(Integer)
            }
        )

    def create_schema(self):
        """
        helper function to create a schema using sqlalchemy orm
        """
        if not self.engine.dialect.has_schema(
            self.engine, 
            self.settings.other.db_schema
        ):
            self.engine.execute(CreateSchema(self.settings.other.db_schema))

    def reset_all_tables(self):
        """
        deletes all tables and creates all tables
        """
        self.Base.metadata.drop_all(self.engine)
        self.Base.metadata.create_all(self.engine, checkfirst=True)

    def reset_tables(self):
        self.create_schema()
        self.reset_all_tables()


from oagdedupe.base import BaseCluster
from oagdedupe.settings import Settings
from oagdedupe.db.database import DatabaseORM
from oagdedupe import utils as du

from dataclasses import dataclass
import networkx as nx
import pandas as pd


@dataclass
class ConnectedComponents(BaseCluster, DatabaseORM):
    """
    Uses a graph to retrieve connected components
    """
    settings:Settings

    def __post_init__(self):
        self.orm = DatabaseORM(settings=self.settings)

    @du.recordlinkage
    def get_df_cluster(self, matches, scores, rl=""):
        """ 
        Convert connected components to dataframe for user friendly output

        Parameters
        ----------
        matches: np.array
            array containing pairs of indices that were predicted matches
        scores: np.array
            vector of match scores

        Returns
        ----------
        pd.DataFrame
            clusters merged with raw data
        """

        df_clusters = getattr(self,f"get_connected_components{rl}")(matches, scores)

        # reset table
        self.engine.execute(f"""
            TRUNCATE TABLE {self.settings.other.db_schema}.clusters;
        """)

        self.orm.bulk_insert(
            df=df_clusters, to_table=self.Clusters
        )
        
        return getattr(self.orm,f"get_clusters{rl}")()

    def get_connected_components(self, matches, scores) -> pd.DataFrame:
        """ 
        Build graph with "matched" candidate pairs, weighted by p(match).
        
        Need to add feature to consider weights when generating 
        connected components.

        Parameters
        ----------
        matches: np.array
            array containing pairs of indices that were predicted matches
        scores: np.array
            vector of match scores

        Returns
        ----------
        pd.DataFrame
            dataframe mapping cluster index to entity index
        """
        g = nx.Graph()
        g.add_weighted_edges_from([
            tuple([f"{match[0]}", f"{match[1]}", score]) 
            for match, score in zip(matches, scores)
        ])
        conn_comp = list(nx.connected_components(g))
        clusters = [
            {"cluster": clusteridx, "_index": int(rec_id), "_type":None}
            for clusteridx, cluster in enumerate(conn_comp)
            for rec_id in cluster
        ]
        return pd.DataFrame(clusters)

    def get_connected_components_link(self, matches, scores) -> pd.DataFrame:
        g = nx.Graph()
        g.add_weighted_edges_from([
            tuple([f"{match[0]}_l", f"{match[1]}_r", score]) 
            for match, score in zip(matches, scores)
        ])
        conn_comp = list(nx.connected_components(g))
        clusters = [
            {
                "cluster": clusteridx, 
                "_index": rec_id.strip("_")[0], 
                "_type": "_l" in rec_id
            }
            for clusteridx, cluster in enumerate(conn_comp)
            for rec_id in cluster
        ]
        return pd.DataFrame(clusters)




# following https://pydantic-docs.helpmanual.io/usage/settings/

# import configparser
# import os
import logging
from typing import Any, List, Optional
from pydantic import BaseSettings, BaseModel
from pathlib import Path


class SettingsService(BaseModel):
    """settings for a service"""

    host: str = "http://172.22.39.26"

    port: int = 8089

    @property
    def url(self) -> str:
        return f"{self.host}:{self.port}"


class SettingsLabelStudio(SettingsService):

    api_key: str = "please provide an api key"

    """project description"""
    description: str = "entity resolution"


class SettingsOther(BaseModel):
    """Other project settings"""

    """dedupe vs record-linkage"""
    dedupe: bool = True

    """block learner sample size (per learning loop)"""
    n: int = 5000

    """maximum length of blocking scheme conjunctions"""
    k: int = 3

    """maximum number of comparisons"""
    max_compare: int = 1_000_000

    """entity attribute names"""
    attributes: Optional[List[str]] = ["name", "addr"]

    """number of cpus to use"""
    cpus: int = 1

    """path to database"""
    path_database: str = "postgresql+psycopg2://username:password@172.22.39.26:8000/db"

    """database schema"""
    db_schema: str = "dedupe"

    """path to model"""
    path_model: Path = Path("./.dedupe/model")

    """label studio settings"""
    label_studio: SettingsLabelStudio = SettingsLabelStudio()

    """fast api settings"""
    fast_api: SettingsService = SettingsService(port=8090)


DEFAULT: SettingsOther = SettingsOther()


class Settings(BaseSettings):
    """project settings"""

    """name of the project, a unique identifier"""
    name: str = "default"

    """path to folder to store the config file"""
    folder: Path = Path("./.dedupe")

    """other project settings"""
    other: Optional[SettingsOther] = SettingsOther()

    class Config:
        env_prefix = "deduper_"

    @property
    def path(self) -> Path:
        return self.folder / f"deduper_settings_{self.name}.json"

    def save(self) -> None:
        """save settings to file"""

        logging.info(f"saving settings for {self.name} to {self.path}")

        # create the settings folder if it doesn't exist
        if not self.folder.is_dir():
            self.folder.mkdir()

        with open(self.path, "w") as f:
            f.write(self.json(indent=2))

    def read(self) -> None:
        """read settings from file"""

        logging.info(f"reading settings for {self.name} from {self.path}")
        self.other = Settings.parse_file(self.path).other

    def sync(self) -> None:
        """read settings if settings file exists and default settings provided, else save

        assume that if default settings are provided and a settings file exists the user wants to read settings from file, not reset settings to default
        """

        if self.other == SettingsOther() and self.path.is_file():
            self.read()
        else:
            self.save()

    def set(self, name: str, value: Any):  # -> Settings:
        """set a setting in self.other and save settings to keep the file updated

        Parameters
        ----------
        name : str
            name of the SettingsOther attribute to set
        value : Any
            value to set to

        Returns
        -------
        Settings
            self after setting, to allow for composed calls to `set` (i.e. settings.set().set()...)
        """
        self.other.__setattr__(name, value)
        self.save()
        return self


def get_settings_from_env() -> Settings:
    """Get settings from environment variables, see https://pydantic-docs.helpmanual.io/usage/settings/

    Returns
    -------
    Settings
        the settings as read from file whose name and location are read from environment variables
    """
    settings = Settings()
    settings.read()
    return settings



from oagdedupe.settings import Settings
from oagdedupe.distance.string import RayAllJaro
from oagdedupe.db.tables import Tables
from sqlalchemy import select, delete, func
from oagdedupe import utils as du

import itertools
from dataclasses import dataclass
import logging

@dataclass
class Initialize(Tables):
    """
    Object used to initialize SQL tables using sqlalchemy

    Can be used to create:
        - df
        - pos/neg
            - created to help build train and labels; 
            - pos contains a random sample repeated 4 times
            - neg contains 10 random samples
        - unlabelled
            - random sample of df of size settings.other.n, 
            - samples are drawn each active learning loop
        - train
            - combines pos, neg, and unlabelled
        - labels
            - gets all distinct pairwise comparisons from train
            - pairs from pos are labelled as a match
            - pairs from neg are labelled as a non-match
    """
    settings:Settings

    def _init_df(self, df, rl=""):
        logging.info(f"Building table {self.settings.other.db_schema}.df{rl}")
        with self.Session() as session:
            session.bulk_insert_mappings(
                getattr(self,f"maindf{rl}"),
                df.to_dict(orient='records')
            )
            session.commit()

    def _init_pos(self):
        # create pos
        pos = select([self.maindf]).order_by(func.random()).limit(1)
        with self.Session() as session:
            records = session.execute(pos).first()
            for i in range(-3,1):
                table = self.Pos()
                for attr in self.settings.other.attributes + ["_index"]:
                    setattr(table, attr, getattr(records[0], attr))
                if i < 0:
                    setattr(table, "_index", i)
                setattr(table, "labelled", True)
                session.add(table)
            session.commit()

    @du.recordlinkage_repeat
    def _init_neg(self, rl=""):
        # create neg
        neg = (
            select([getattr(self,f"maindf{rl}")]).order_by(func.random()).limit(10)
        )
        with self.Session() as session:
            records = session.execute(neg).all()
            for r in records:
                table = getattr(self,f"Neg{rl}")()
                for attr in self.settings.other.attributes + ["_index"]:
                    setattr(table, attr, getattr(r[0], attr))
                setattr(table, "labelled", True)
                session.add(table)
            session.commit()

    @du.recordlinkage_repeat
    def _init_unlabelled(self, rl=""):
        # create unlabelled
        unlabelled = (
            select([getattr(self, f"maindf{rl}")])
            .order_by(func.random())
            .limit(self.settings.other.n)
        )
        with self.Session() as session:
            records = session.execute(unlabelled).all()
            for r in records:
                table = getattr(self,f"Unlabelled{rl}")()
                for attr in self.settings.other.attributes + ["_index"]:
                    setattr(table, attr, getattr(r[0], attr))
                setattr(table, "labelled", False)
                session.add(table)
            session.commit()

    @du.recordlinkage_repeat
    def _init_train(self, rl=""):
        
        logging.info(f"Building table {self.settings.other.db_schema}.train{rl}")        
        # create train
        with self.Session() as session:
            for tab in [getattr(self,f"Unlabelled{rl}"), self.Pos, getattr(self,f"Neg{rl}")]:
                records = session.query(tab).all()
                for r in records:
                    train = getattr(self,f"Train{rl}")()
                    for attr in self.settings.other.attributes + ["_index", "labelled"]:
                        setattr(train, attr, getattr(r, attr))
                    session.merge(train)
            session.commit()

    def _init_labels(self):
        logging.info(f"Building table {self.settings.other.db_schema}.labels.")
        with self.Session() as session:
            for l,tab in [(1,self.Pos), (0,self.Neg)]:
                records = session.query(tab).all()
                pairs = list(itertools.combinations(records, 2))
                for left,right in pairs:
                    label = self.Labels()
                    if left._index < right._index:
                        label._index_l = left._index
                        label._index_r = right._index
                        label.label = l
                        session.add(label)
            session.commit()
        self._label_distances()

    def _init_labels_link(self):
        logging.info(f"Building table {self.settings.other.db_schema}.labels_link.")
        with self.Session() as session:
            for l,tab,tab_link in [(1,self.Pos,self.Pos), (0,self.Neg,self.Neg_link)]:
                records = session.query(tab).all()
                records_link = session.query(tab_link).all()
                for left,right in zip(records, records_link):
                    label = self.Labels()
                    label._index_l = left._index
                    label._index_r = right._index
                    label.label = l
                    session.add(label)
            session.commit()
        self._label_distances()

    def _label_distances(self):
        """
        computes distances between pairs of records from labels table 
        then appends distances to the labels table
        """
        self.distance = RayAllJaro(settings=self.settings)
        self.distance.save_distances(
            table=self.Labels,
            newtable=self.Labels
        )

    @du.recordlinkage_repeat
    def _resample(self, rl=""):

        # delete unlabelled from train
        with self.Session() as session:
            stmt = (
                delete(getattr(self,f"Train{rl}")).
                where(getattr(self,f"Train{rl}").labelled==False)
            )
            session.execute(stmt)
            session.commit()

        # resample unlabelled
        self.engine.execute(f"""
                TRUNCATE TABLE {self.settings.other.db_schema}.unlabelled{rl};
            """)

        # add to train
        with self.Session() as session:
     
            records = session.query(getattr(self,f"Unlabelled{rl}")).all()
            for r in records:
                train = getattr(self,f"Train{rl}")()
                for attr in self.settings.other.attributes + ["_index", "labelled"]:
                    setattr(train, attr, getattr(r, attr))
                session.merge(train)
            session.commit()

    @du.recordlinkage
    def setup(self, df=None, df2=None, reset=True, resample=False, rl=""):
        """
        runs table creation functions

        Parameters
        ----------
        df: Optional[pd.DataFrame]
            dataframe to dedupe
        reset: bool
            set True to delete and create all tables
        resample: bool
            used for active learning loops where model needs to pull a new 
            sample, without deleting df, train, or labels
        """
        
        # initialize Tables sqlalchemy classes
        self.setup_dynamic_declarative_mapping()

        if reset:
            self.reset_tables()

            logging.info(f"building tables in schema: {self.settings.other.db_schema}")
            if "_index" in df.columns:
                raise ValueError("_index cannot be a column name")
            
            self._init_df(df=df)
            if self.settings.other.dedupe == False:
                self._init_df(df=df2, rl="_link")

            self._init_pos()
            self._init_neg()
            self._init_unlabelled()
            self._init_train()

            getattr(self, f"_init_labels{rl}")()

        if resample:
            self._resample()
            self._label_distances()

    


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['oagdedupe',
 'oagdedupe.block',
 'oagdedupe.cluster',
 'oagdedupe.db',
 'oagdedupe.distance',
 'oagdedupe.fastapi',
 'oagdedupe.labelstudio',
 'oagdedupe.postgres']

package_data = \
{'': ['*']}

install_requires = \
['Faker>=13.15.1,<14.0.0',
 'SQLAlchemy>=1.4.39,<2.0.0',
 'Sphinx>=5.1.1,<6.0.0,!=5.2.0.post0',
 'diagrams>=0.21.1,<0.22.0',
 'fastapi[all]>=0.79.0,<0.80.0',
 'flake8>=4.0.1,<5.0.0',
 'graphviz>=0.19.0,<0.20.0',
 'ipykernel>=6.13.0,<7.0.0',
 'jellyfish>=0.9.0,<0.10.0',
 'matplotlib>=3.5.1,<4.0.0',
 'modAL>=0.4.1,<0.5.0',
 'myst-parser>=0.18.0,<0.19.0',
 'nbconvert>=6.5.1,<7.0.0',
 'networkx>=2.8,<3.0',
 'numpy>=1.22.1,<2.0.0',
 'pandas>=1.4.2,<2.0.0',
 'pathos>=0.2.9,<0.3.0',
 'protobuf>=3.20.2,<4.0.0',
 'psycopg2-binary>=2.9.3,<3.0.0',
 'pydantic>=1.10.2,<2.0.0',
 'pytest>=7.1.2,<8.0.0',
 'ray>=1.13.0,<2.0.0',
 'scikit-learn>=1.0.2,<2.0.0',
 'seaborn>=0.11.2,<0.12.0',
 'sphinx-rtd-theme>=1.0.0,<2.0.0',
 'streamlit-aggrid>=0.2.3,<0.3.0',
 'streamlit>=1.11.1,<2.0.0',
 'tqdm>=4.58.0,<5.0.0']

extras_require = \
{'book': ['jupytext>=1.14.1,<2.0.0', 'autodocsumm>=0.2.9,<0.3.0']}

setup_kwargs = {
    'name': 'oagdedupe',
    'version': '0.1.0',
    'description': 'oagdedupe is a Python library for scalable entity resolution, using active learning to learn blocking configurations, generate comparison pairs, then clasify matches.',
    'long_description': '# oagdedupe  \n\noagdedupe is a Python library for scalable entity resolution, using active \nlearning to learn blocking configurations, generate comparison pairs, \nthen clasify matches. \n\n## page contents\n- [Documentation](#documentation)\n- [Installation](#installation)\n    - [label-studio](#label-studio)\n    - [postgres](#postgres)\n    - [project settings](#project-settings)\n- [dedupe](#dedupe-example)\n- [record-linkage](#record-linkage-example)\n    \n# Documentation<a name="#documentation"></a>\n\nYou can find the documentation of oagdedupe at https://deduper.readthedocs.io/en/latest/, \nwhere you can find the [api reference](https://deduper.readthedocs.io/en/latest/dedupe/api.html), \n[guide to methodology](https://deduper.readthedocs.io/en/latest/userguide/intro.html),\nand [examples](https://deduper.readthedocs.io/en/latest/examples/example_dedupe.html).\n\n# Installation<a name="#installation"></a>\n\n[tbd pip install instructions]\n\n## start label-studio<a name="#label-studio"></a>\n\nStart label-studio using docker command below, updating `[LS_PORT]` to the \nport on your host machine\n\n```\ndocker run -it -p [LS_PORT]:8080 -v `pwd`/cache/mydata:/label-studio/data \\\n\t--env LABEL_STUDIO_LOCAL_FILES_SERVING_ENABLED=true \\\n\t--env LABEL_STUDIO_LOCAL_FILES_DOCUMENT_ROOT=/label-studio/files \\\n\t-v `pwd`/cache/myfiles:/label-studio/files \\\n\theartexlabs/label-studio:latest label-studio\n```\n\n## postgres<a name="#postgres"></a>\n\n[insert instructions here about initializing postgres]\n\nmost importantly, need to create functions (dedupe/postgres/funcs.py)\n\n## project settings<a name="#project-settings"></a>\n\nMake a `dedupe.settings.Settings` object. For example:\n```py\nfrom oagdedupe.settings import (\n    Settings,\n    SettingsOther,\n)\n\nsettings = Settings(\n    name="default",  # the name of the project, a unique identifier\n    folder="./.dedupe",  # path to folder where settings and data will be saved\n    other=SettingsOther(\n        n=5000, # active-learning samples per learning loop\n        k=3, # max_len of block conjunctions\n        cpus=20,  # parallelize distance computations\n        attributes=["givenname", "surname", "suburb", "postcode"],  # list of entity attribute names\n        path_database="postgresql+psycopg2://username:password@172.22.39.26:8000/db",  # where to save the sqlite database holding intermediate data\n        db_schema="dedupe",\n        path_model="./.dedupe/test_model",  # where to save the model\n        label_studio={\n            "port": 8089,  # label studio port\n            "api_key": "83e2bc3da92741aa41c272829558c596faefa745",  # label studio port\n            "description": "chansoo test project",  # label studio description of project\n        },\n        fast_api={"port": 8090},  # fast api port\n    ),\n)\nsettings.save()\n```\nTo get label studio api_key:\n   1. log in (can make up any user/pw).\n   2. Go to "Account & Settings" using icon on top-right\n   3. Get Access Token and copy/paste into settings at `settings.other.label_studio["api_key"]` \n\nSee [dedupe/settings.py](./dedupe/settings.py) for the full settings code.\n\n# dedupe<a name="#dedupe-example"></a>\n\nBelow is an example that dedupes `df` on attributes columns specified in settings.\n\n## train dedupe<a name="#train-dedupe"></a>\n\n```py\nimport glob\nimport pandas as pd\nfrom oagdedupe.api import Dedupe\n\nd = Dedupe(settings=settings)\nd.initialize(df=df, reset=True)\n\n# %%\n# pre-processes data and stores pre-processed data, comparisons, ID matrices in SQLite db\nd.fit_blocks()\n```\n\n# record-linkage<a name="#record-linkage-example"></a>\n\nBelow is an example that links `df` to `df2`, on attributes columns specified \nin settings (dataframes should share these columns).\n\n## train record-linkage<a name="#train-record-linkage"></a>\n```py\nimport glob\nimport pandas as pd\nfrom oagdedupe.api import RecordLinkage\n\nd = RecordLinkage(settings=settings)\nd.initialize(df=df, df2=df2, reset=True)\n\n# %%\n# pre-processes data and stores pre-processed data, comparisons, ID matrices in SQLite db\nd.fit_blocks()\n```\n\n# active learn<a name="#active-learn"></a>\n\nFor either dedupe or record-linkage, run:\n\n```sh\n   DEDUPER_NAME="<project name>";\n   DEDUPER_FOLDER="<project folder>";\n   python -m dedupe.fastapi.main\n```\n\nreplacing `<project name>` and `<project folder>` with your project settings (for the example above, `test` and `./.dedupe`).\n\nThen return to label-studio and start labelling. When the queue falls under 5 tasks, fastAPI will update the model with labelled samples then send more tasks to review.\n\n# predictions<a name="#predictions"></a>\n\nTo get predictions, simply run the `predict()` method.\n\nDedupe:\n```py\nd = Dedupe(settings=Settings(name="test", folder="./.dedupe"))\nd.predict()\n```\n\nRecord-linkage:\n```py\nd = RecordLinkage(settings=Settings(name="test", folder="./.dedupe"))\nd.predict()\n```',
    'author': 'Chansoo Song',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/chansooligans/oagdedupe',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



from oagdedupe.db.database import DatabaseORM,DatabaseCore
from oagdedupe.settings import Settings
from oagdedupe.block import Blocker
from oagdedupe import utils as du

from dataclasses import dataclass
from typing import List
from functools import lru_cache, cached_property
import pandas as pd
import itertools
from multiprocessing import Pool
import logging
import tqdm

class InvertedIndex:
    """
    Used to build inverted index. An inverted index is dataframe 
    where keys are signatures and values are arrays of entity IDs
    """

    @du.recordlinkage
    def get_pairs(self, names, table, rl=""):
        """
        Given forward index, construct inverted index. 
        Then for each row in inverted index, get all "nC2" distinct 
        combinations of size 2 from the array. Concatenates and 
        returns all distinct pairs.

        Parameters
        ----------
        names : List[str]
            list of block schemes
        table : str
            table name of forward index

        Returns
        ----------
        pd.DataFrame
        """
        return getattr(self.db, f"get_inverted_index_pairs{rl}")(
            names=names,
            table=table
        )

    def get_pairs_in_memory(self, names, table="blocks_train"):
        """
        same as get_pairs() but runs in-memory;
        faster and used when searching for best conjunctions

        Parameters
        ----------
        names : List[str]
            list of block schemes
        table : str
            table name of forward index

        Returns
        ----------
        pd.Dataframe
        """
        inverted_index = self.db.get_inverted_index(names,table)
        return pd.DataFrame([ 
                y
                for x in list(inverted_index["array_agg"])
                for y in list(itertools.combinations(x, 2))
            ], columns = ["_index_l","_index_r"]
        ).assign(blocked=True).drop_duplicates()

    def get_pairs_in_memory_link(self, names, table="blocks_train"):
        """
        same as get_pairs() but runs in-memory;
        faster and used when searching for best conjunctions

        Parameters
        ----------
        names : List[str]
            list of block schemes
        table : str
            table name of forward index

        Returns
        ----------
        pd.Dataframe
        """
        inverted_index = self.db.get_inverted_index(names,table)
        inverted_index_link = self.db.get_inverted_index(names,f"{table}_link")
        return (
            (
                inverted_index
                .explode("array_agg")
                .rename({"array_agg":"_index_l"}, axis=1)
            )
            .merge(
                inverted_index_link
                .explode("array_agg")
                .rename({"array_agg":"_index_r"}, axis=1)
            )
            [["_index_l","_index_r"]]
            .assign(blocked=True).drop_duplicates()
        )
        

class DynamicProgram(InvertedIndex):
    """
    Given a block scheme, use dynamic programming algorithm getBest()
    to construct best conjunction
    """

    def get_stats(self, names, n_pairs, coverage):
        """
        Evaluate conjunction performance by:
            - percentage of positive / negative labels that are blocked;
            uses blocks_train
            - get comparisons using train data to compute reduction ratio;
            uses blocks_train

        Parameters
        ----------
        names: List[str]
            list of block schemes
        n_pairs: int
            number of comparison pairs obtained from train data
        coverage: pd.DataFrame
            labelled data merged with comparison pairs obtained from 
            train data; contains binary column blocked, which is 1 if 
            the labelled pair was blocked and 0 if it was not blocked;
            used to compute percent of positive and negative labels covered

        Returns
        ----------
        dict
            returns statistics evaluating the conjunction: 
                - reduction ratio
                - percent of positive pairs blocked
                - percent of negative pairs blocked
                - number of pairs generated
                - length of conjunction
        """
        return {
            "scheme": names,
            "rr":1 - (n_pairs / (self.db.n_comparisons)),
            "positives":coverage.loc[coverage["label"]==1, "blocked"].mean(),
            "negatives":coverage.loc[coverage["label"]==0, "blocked"].mean(),
            "n_pairs": n_pairs,
            "n_scheme": len(names)
        }

    @du.recordlinkage
    def get_coverage(self, names, rl=""):
        """
        Get comparisons using train data then merge with labelled data. 

        Parameters
        ----------
        names: List[str]
            list of block schemes

        Returns
        ----------
        dict
            returns statistics evaluating the conjunction: 
                - reduction ratio
                - percent of positive pairs blocked
                - percent of negative pairs blocked
                - number of pairs generated
                - length of conjunction
        """

        
        train_pairs = getattr(self,f"get_pairs_in_memory{rl}")(names=names)
        
        coverage = self.db.get_labels().merge(train_pairs, how = 'left')
        coverage = coverage.fillna(0)
        
        return self.get_stats(
            names=names, n_pairs=len(train_pairs), coverage=coverage
        )

    @lru_cache
    def score(self, arr:tuple):
        """
        Wraps get_coverage() function with @lru_cache decorator for caching.

        Parameters
        ----------
        arr: tuple
            tuple of block schemes
        """
        return self.get_coverage(names=list(arr))

    def getBest(self, scheme:tuple):
        """
        Dynamic programming implementation to get best conjunction.

        Parameters
        ----------
        scheme: tuple
            tuple of block schemes
        """

        dp = [None for _ in range(self.settings.other.k)]
        dp[0] = self.score(scheme)

        if (dp[0]["positives"] == 0) or (dp[0]["rr"] < 0.99) or (dp[0]["rr"] == 1):
            return None

        for n in range(1,self.settings.other.k):

            scores = [
                self.score(
                    tuple(sorted(dp[n-1]["scheme"] + [x]))
                ) 
                for x in self.db.blocking_schemes
                if x not in dp[n-1]["scheme"]
            ]

            scores = [
                x
                for x in scores
                if (x["positives"] > 0) & \
                    (x["rr"] < 1) & (x["n_pairs"] > 1) & \
                    (sum(["_ngrams" in _ for _ in x["scheme"]]) <= 1)
            ]
            
            if len(scores) == 0:
                return dp[:n]

            dp[n] = max(
                scores, 
                key=lambda x: (
                    x["rr"], 
                    x["positives"], 
                    -x["negatives"],
                    -x["n_scheme"]
                )
            )

        return dp

class Conjunctions(DynamicProgram):
    """
    For each block scheme, get the best block scheme conjunctions of 
    lengths 1 to k using greedy dynamic programming approach.
    """

    def __init__(self, settings:Settings):
        self.settings = settings
        self.db = DatabaseCore(settings=self.settings) 

    @property
    def conjunctions(self):
        """
        Computes conjunctions for each block scheme in parallel
        """
        logging.info("note:due to caching, iterations get progressively faster")
        schemes = [tuple([o]) for o in self.db.blocking_schemes]
        with Pool(self.settings.other.cpus) as p:
            res = list(tqdm.tqdm(
                p.imap(
                    self.getBest, 
                    schemes
                ), 
                total=len(schemes)
            ))
        
        return res
    
    @cached_property
    def df_conjunctions(self):
        """
        DataFrame containing best conjunctions and their stats
        """
        logging.info(f"getting best conjunctions")        
        df = pd.concat([
            pd.DataFrame(r)
            for r in self.conjunctions
        ]).reset_index(drop=True)
        return df.loc[
            df.astype(str).drop_duplicates().index
        ].sort_values("rr", ascending=False)

    @property
    def newtablemap(self):
        return {
            "comparisons":self.orm.Comparisons,
            "full_comparisons":self.orm.FullComparisons
        }

    def get_comparisons(
            self, 
            table, 
            n_covered
        ):
        """
        Applies subset of best conjunction to obtain comparison pairs.

        Parameters
        ----------
        table: str
            get pairs from table (either blocks_train for sample or 
            blocks_df for full df)
        n_covered: int
            number of records that the conjunctions should cover
        """
        df = pd.DataFrame()
        
        self.blocker = Blocker(settings=self.settings)

        for stats in self.df_conjunctions.to_dict(orient="records"):

            logging.info(f"""evaluating scheme {stats["scheme"]}""")

            if stats["rr"] < self.db.min_rr:
                logging.warning(f"""
                    next conjunction exceeds reduction ratio limit;
                    stopping pair generation with scheme {stats["scheme"]}
                """)
                return df

            if table == "blocks_df":
                self.blocker.build_forward_indices_full(stats["scheme"])
                
            df = pd.concat([
                df,
                self.get_pairs(names=stats["scheme"], table=table)
            ], axis=0).drop_duplicates()

            logging.info(f"""{len(df)} comparison pairs gathered""")

            if len(df) > n_covered:
                return df.drop(["blocked"], axis=1)
        
    def save_best(
        self, 
        table, 
        newtable,
        n_covered
    ):
        """
        Applies subset of best conjunction to obtain comparison pairs.

        Parameters
        ----------
        table: str
            get pairs from table (either blocks_train for sample or 
            blocks_df for full df)
        newtable: str
            table to save output (either comparisons for sample or 
            full_comparisons for full df)
        n_covered: int
            number of records that the conjunctions should cover
        """
        
        comparisons = self.get_comparisons(table=table,n_covered=n_covered)
        
        self.orm = DatabaseORM(settings=self.settings)
        
        self.orm.truncate_table(newtable)
        self.orm.bulk_insert(
            df=comparisons, to_table=self.newtablemap[newtable]
        )


from pydantic import BaseModel
from oagdedupe.settings import Settings

from dataclasses import dataclass
from typing import List, Dict, Optional
import requests
import pandas as pd
import json

class Annotation(BaseModel):
    id: int
    created_username: str
    created_ago: str
    result: List[dict]
    was_cancelled: bool
    ground_truth: bool
    created_at: str
    updated_at: str
    lead_time: float
    task: int
    completed_by: int

    @property
    def label_map(self):
        return {"Match": 1, "Not a Match": 0, "Uncertain": 2}

    @property
    def label(self):
        annotation = self.result[0]["value"]["choices"][0]
        return self.label_map[annotation]

class Task(BaseModel):
    id: int
    cancelled_annotations: int
    total_annotations: int
    total_predictions: int
    updated_by: List[dict]
    data: dict
    created_at: str
    updated_at: str
    is_labeled: bool
    overlap: int 
    project: int

class TaskList(BaseModel):
    total_annotations: int
    total: int
    tasks: List[Task]

    @property
    def n_incomplete(self):
        return self.total - self.total_annotations

class Project(BaseModel):
    id: int
    title: str
    description: Optional[str]
    label_config: str
    created_at: str
    created_by: dict
    num_tasks_with_annotations: Optional[int]
    task_number: Optional[int]

@dataclass
class APIProjects:

    def list_projects(self):
        content = requests.get(
            f"{self.url}/api/projects", 
            headers=self.headers
        )
        return [
            Project.parse_obj(p)
            for p in json.loads(content.content)["results"]
        ]

    def create_project(
        self, title: str = "new project", description: str = "description"
    ):

        new_project = {
            "title": title,
            "description": description,
            "label_config": """
                <View>
                    <Header value="Table with {key: value} pairs"/>
                    <Table name="table" value="$item"/>
                    <Choices name="choice" toName="table">
                        <Choice value="Match"/>
                        <Choice value="Not a Match"/>
                        <Choice value="Uncertain"/>
                    </Choices>
                </View>
            """,
        }

        resp = requests.post(
            f"{self.url}/api/projects", headers=self.headers, data=new_project
        )

        return Project.parse_obj(json.loads(resp.content))

@dataclass
class APITasks:

    def get_tasks(self, project_id):
        tasklist = json.loads(
            requests.get(
                f"{self.url}/api/tasks", 
                headers=self.headers, 
                data={"project": project_id}
            ).content
        )
        tasklist["tasks"] = [Task.parse_obj(t) for t in tasklist["tasks"]]
        return TaskList.parse_obj(tasklist)

    def post_tasks(self, df, project_id):
        for _, row in df.iterrows():
            resp = requests.post(
                f"{self.url}/api/tasks", 
                headers=self.headers, 
                data={
                    "project": project_id, 
                    "data": json.dumps({"item": row.to_dict()})
                }
            )
            json.loads(resp.content)


@dataclass
class APIAnnotations:

    def _get_annotationlist_for_task(self, task_id):
        annotationlist = json.loads(
            requests.get(
                f"{self.url}/api/tasks/{task_id}/annotations/", 
                headers=self.headers
            ).content
        )
        if annotationlist:
            annotationlist = [
                Annotation.parse_obj(a)
                for a in annotationlist
            ]
        return annotationlist

    def _latest_annotation(self, annotationlist):
        return sorted(
            annotationlist, 
            key=lambda d: d.created_at
        )[0].label

    def get_new_labels(self, project_id):
        labels = {}
        for task in self.get_tasks(project_id=project_id).tasks:
            annotationlist = self._get_annotationlist_for_task(task_id=task.id)
            if annotationlist:
                labels[task.id] = {
                    **{"label":self._latest_annotation(annotationlist)},
                    **task.data["item"]
                }
        return pd.DataFrame(labels).T

@dataclass
class APIWebhooks:
    settings: Settings

    def get_webhooks(self):
        resp = requests.get(f"{self.url}/api/webhooks", headers=self.headers)
        return json.loads(resp.content)

    def post_webhook(self, project_id):
        query = {
            "project": project_id,
            "url": f"{self.settings.other.fast_api.url}/payload",
            "send_payload": False,
            "is_active": True,
            "actions": ["ANNOTATION_CREATED", "ANNOTATION_UPDATED"],
        }
        resp = requests.post(
            f"{self.url}/api/webhooks", headers=self.headers, data=query
        )


@dataclass
class LabelStudioAPI(APIProjects, APITasks, APIAnnotations, APIWebhooks):
    settings: Settings

    @property
    def url(self) -> str:
        return self.settings.other.label_studio.url

    @property
    def headers(self) -> Dict[str, str]:
        return {
            "Authorization": f"""Token {self.settings.other.label_studio.api_key}"""
        }


from oagdedupe.settings import Settings
from oagdedupe.db.tables import Tables
from oagdedupe import utils as du

from functools import cached_property
import sqlalchemy
from sqlalchemy import create_engine
from sqlalchemy.orm import aliased
from dataclasses import dataclass
import pandas as pd
import numpy as np

def check_unnest(name):
    if "ngrams" in name:
        return f"unnest({name})"
    return name

def signatures(names):
    return ", ".join([
        f"{check_unnest(name)} as signature{i}" 
        for i,name in  enumerate(names)
    ])

@dataclass
class DatabaseCore:
    """
    Object contains methods to query database using sqlalchemy CORE;
    It's easier to use sqlalchemy core than ORM for parallel operations.
    """
    settings: Settings

    def query(self, sql):
        """
        for parallel implementation, need to create separate engine 
        for each process
        """
        engine = create_engine(self.settings.other.path_database)
        res = pd.read_sql(sql, con=engine)
        engine.dispose()
        return res

    def get_labels(self):
        """
        query the labels table

        Returns
        ----------
        pd.DataFrame
        """
        return self.query(
            f"""
            SELECT * FROM {self.settings.other.db_schema}.labels
            """
        )

    def get_inverted_index(self, names, table):
        """
        see dedupe.block.learner.InvertedIndex;

        Given forward index, construct inverted index. 
        Then for each row in inverted index, get all "nC2" distinct 
        combinations of size 2 from the array. 
        
        Returns inverted Index

        Parameters
        ----------
        names : List[str]
            list of block schemes
        table : str
            table name of forward index

        Returns
        ----------
        pd.DataFrame
        """
        return self.query(
            f"""
            WITH 
                inverted_index AS (
                    SELECT 
                        {signatures(names)}, 
                        ARRAY_AGG(_index ORDER BY _index asc) as array_agg
                    FROM {self.settings.other.db_schema}.{table}
                    GROUP BY {", ".join([f"signature{i}" for i in range(len(names))])}
                )
            SELECT * 
            FROM inverted_index
            WHERE array_length(array_agg, 1) > 1
            """
        )

    def get_inverted_index_pairs(self, names, table):
        """
        see dedupe.block.learner.InvertedIndex;

        Given forward index, construct inverted index. 
        Then for each row in inverted index, get all "nC2" distinct 
        combinations of size 2 from the array. 
        
        Concatenates and returns all distinct pairs.

        Parameters
        ----------
        names : List[str]
            list of block schemes
        table : str
            table name of forward index

        Returns
        ----------
        pd.DataFrame
        """
        return self.query(f"""
            WITH 
                inverted_index AS (
                    SELECT 
                        {signatures(names)}, 
                        ARRAY_AGG(_index ORDER BY _index asc) as array_agg
                    FROM {self.settings.other.db_schema}.{table}
                    GROUP BY {", ".join([f"signature{i}" for i in range(len(names))])}
                ),
                inverted_index_subset AS (
                    SELECT unnest_2d_1d(combinations(array_agg)) as pairs
                    FROM inverted_index
                    WHERE array_length(array_agg, 1) > 1
                )
            SELECT pairs[1] as _index_l, pairs[2] as _index_r, True as blocked
            FROM inverted_index_subset
            """)

    def get_inverted_index_pairs_link(self, names, table):
        """
        see dedupe.block.learner.InvertedIndex;

        Given forward index, construct inverted index. 
        Then for each row in inverted index, get all "nC2" distinct 
        combinations of size 2 from the array. 
        
        Concatenates and returns all distinct pairs.

        Parameters
        ----------
        names : List[str]
            list of block schemes
        table : str
            table name of forward index

        Returns
        ----------
        pd.DataFrame
        """
        aliases = [f"signature{i}" for i in range(len(names))]
        return self.query(f"""
            WITH 
                inverted_index AS (
                    SELECT 
                        {signatures(names)}, 
                        unnest(ARRAY_AGG(_index ORDER BY _index asc)) _index_l
                    FROM {self.settings.other.db_schema}.{table}
                    GROUP BY {", ".join(aliases)}
                ),
                inverted_index_link AS (
                    SELECT 
                        {signatures(names)}, 
                        unnest(ARRAY_AGG(_index ORDER BY _index asc)) _index_r
                    FROM {self.settings.other.db_schema}.{table}_link
                    GROUP BY {", ".join(aliases)}
                )
            SELECT _index_l, _index_r, True as blocked
            FROM inverted_index t1
            JOIN inverted_index_link t2
                ON {" and ".join(
                    [f"t1.{s} = t2.{s}" for s in aliases]
                )}
            """)

    @cached_property
    def blocking_schemes(self):
        """
        Get all blocking schemes

        Returns
        ----------
        List[str]
        """
        return self.query(
            f"SELECT * FROM {self.settings.other.db_schema}.blocks_train LIMIT 1"
        ).columns[1:].tolist()

    @du.recordlinkage_both
    def n(self, rl=""):
        return self.query(
            f"SELECT count(*) FROM {self.settings.other.db_schema}.df{rl}"
        )["count"].values[0]

    @property
    def n_comparisons(self):
        """number of total possible comparisons"""
        n = self.n()
        if self.settings.other.dedupe == False:
            return np.product(n)
        return (n * (n-1))/2
    
    @property
    def min_rr(self):
        """minimum reduction ratio"""
        reduced = (self.n_comparisons - self.settings.other.max_compare) 
        return reduced / self.n_comparisons

@dataclass
class Engine:
    """
    manages non-ORM textual connections to database
    """
    settings: Settings

    @cached_property
    def engine(self):
        return create_engine(self.settings.other.path_database)

@dataclass
class DatabaseORM(Tables, DatabaseCore, Engine):
    """
    Object to query database using sqlalchemy ORM. 
    Uses the Session object as interface to the database.
    """
    settings: Settings

    def get_train(self):
        """
        query the train table

        Returns
        ----------
        pd.DataFrame
        """
        with self.Session() as session:
            query = session.query(self.Train)
            return pd.read_sql(query.statement, query.session.bind)

    def get_distances(self):
        """
        query unlabelled distances for sample data

        Returns
        ----------
        pd.DataFrame
        """
        with self.Session() as session:
            query = (
                session
                .query(self.Distances)
                .join(
                    self.Labels, 
                    (self.Distances._index_l==self.Labels._index_l) & 
                    (self.Distances._index_r==self.Labels._index_r), 
                    isouter=True)
                .filter(self.Labels.label == None)
                )
            return pd.read_sql(query.statement, query.session.bind)

    def get_full_distances(self):
        """
        query distances for full data

        Returns
        ----------
        pd.DataFrame
        """
        with self.Session() as session:
            query = (
                session
                .query(
                    *(
                        getattr(self.FullDistances,x) 
                        for x in self.settings.other.attributes
                    )
                )
                )
            return pd.read_sql(query.statement, query.session.bind)

    def get_full_comparison_indices(self):
        """
        query indices of comparison pairs for full data

        Returns
        ----------
        pd.DataFrame
        """
        with self.Session() as session:
            query = (
                session
                .query(self.FullDistances._index_l, self.FullDistances._index_r)
                .order_by(self.FullDistances._index_l, self.FullDistances._index_r)
                )
            return pd.read_sql(query.statement, query.session.bind)

    def get_compare_cols(self):
        """
        gets comparison columns with "_l" and "_r" suffices

        Returns
        ----------
        List[str]

        Examples
        ----------
        >>> self.settings.other.attributes = ["name", "address"]
        >>> get_compare_cols()
        ["name_l", "address_l", "index_l", "name_r", "address_r", "index_r"]
        """
        columns = [
            [f"{x}_l" for x in self.settings.other.attributes], 
            ["_index_l"],
            [f"{x}_r" for x in self.settings.other.attributes],
            ["_index_r"]
        ]
        return sum(columns, [])

    @du.recordlinkage
    def fields_table(self, table, rl=""):
        mapping = {
            "comparisons":"Train",
            "full_comparisons":"maindf",
            "labels":"Train"
        }
        
        return (
            aliased(getattr(self,mapping[table])), 
            aliased(getattr(self,mapping[table]+rl))
        )
        
    def labcol(self, table):
        if "comparisons" in table.__tablename__:
            return table._index_l.label("drop")
        return table.label

    def get_comparison_attributes(self, table):
        """
        merge attributes on to dataframe with just comparison pair indices
        assign "_l" and "_r" suffices

        Returns
        ----------
        pd.DataFrame
        """

        with self.Session() as session:
            
            dataL, dataR = self.fields_table(table.__tablename__)
            
            query = (
                session
                .query(
                    self.labcol(table),
                    *(
                        getattr(dataL,x).label(f"{x}_l")
                        for x in self.settings.other.attributes + ["_index"]
                    ),
                    *(
                        getattr(dataR,x).label(f"{x}_r")
                        for x in self.settings.other.attributes + ["_index"]
                    ),
                )
                .outerjoin(dataL, table._index_l==dataL._index)
                .outerjoin(dataR, table._index_r==dataR._index)
                .order_by(table._index_l, table._index_r)
                )

            return (
                pd.read_sql(query.statement, query.session.bind)
                .drop(["drop"], axis=1, errors='ignore')
            )

    def get_clusters(self):
        with self.Session() as session:
            query = (
                session
                .query(self.maindf, self.Clusters.cluster)
                .outerjoin(self.Clusters.cluster, self.Clusters._index == self.maindf._index)
                .order_by(self.Clusters.cluster)
                )
            return pd.read_sql(query.statement, query.session.bind)

    def get_clusters_link(self):
        with self.Session() as session:
            dflist = []
            for _type in [True, False]:
                subquery = session.query(
                    self.Clusters.cluster,self.Clusters._index,self.Clusters._type
                ).filter(self.Clusters._type==_type).subquery()
                query = (
                    session
                    .query(self.maindf_link, subquery.c.cluster)
                    .outerjoin(subquery, subquery.c._index["_index"] == self.maindf_link._index)
                    .order_by(subquery.c.cluster)
                    )
                dflist.append(pd.read_sql(query.statement, query.session.bind))
            return dflist

    def truncate_table(self, table):
        self.engine.execute(f"""
            TRUNCATE TABLE {self.settings.other.db_schema}.{table};
        """)

    def _update_table(self, df, to_table):
        with self.Session() as session:    
            for r in df.to_dict(orient="records"):
                for k in r.keys():
                    setattr(to_table, k, r[k])
                session.merge(to_table)
            session.commit()
        
    def bulk_insert(self, df, to_table):
        with self.Session() as session:
            session.bulk_insert_mappings(
                to_table, 
                df.to_dict(orient='records')
            )
            session.commit()